! WHIZARD 3.1.6 Mar 20 2025
!
! Copyright (C) 1999-2025 by
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!
!     with contributions from
!     cf. main AUTHORS file
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module electron_pdfs

  use kinds, only: default
  use io_units
  use sm_qed, only: alpha_qed_t, alpha_qed_fixed_t, alpha_qed_from_scale_t

  implicit none
  private

  public :: qed_pdf_t
  public :: coeffqed_b0
  public :: coeffqed_b1
  public :: set_qed_pdf_parameters
  public :: elec_asym
  public :: phot_asym
  public :: recasym
  public :: bar_asym
  public :: recnum
  public :: recbar
  public :: rechat
  public :: elecbar_asym_p
  public :: photbar_asym_p
  public :: recbar_singlet
  public :: recbar_nonsinglet
  public :: recbar_photon
  public :: recbar_ele
  public :: recbar_pos
  public :: rechat_singlet
  public :: rechat_nonsinglet
  public :: rechat_photon
  public :: rechat_ele
  public :: rechat_pos
  public :: t_alpha
  public :: full_series
  public :: rec_log_series
  public :: add_logvec
  public :: sub_logvec
  public :: endpoint_func_S
  public :: endpoint_func_NS
  public :: endpoint_func_GAM
  public :: endpoint_func_POS
  public :: endpoint_func_ELE
  public :: elec_pdf

  integer, parameter, public :: EPDF_ELE = 0, EPDF_POS = 1, &
       EPDF_S = 2, EPDF_NS = 3, EPDF_G = 4

  integer, parameter :: wp = default
  real(wp),parameter :: &
       pi     = 3.1415926535897932384626433832795028841972_wp, &
       zeta2  = 1.6449340668482264364724151666460251892189_wp, &
       zeta3  = 1.2020569031595942853997381615114499907649_wp, &
       zeta4  = 1.0823232337111381915160036965411679027747_wp, &
       zeta5  = 1.0369277551433699263313654864570341680570_wp, &
       eulerc = 0.5772156649015328606065120900824024310421_wp, &
       pi2    = pi**2, &
       pi4    = pi**4, &
       ln2    = log(2._wp), &
       ln10   = log(10._wp)

  type :: qed_pdf_t
     private
     integer :: flv = 0
     class(alpha_qed_t), allocatable :: aqed
     real(wp) :: mass = 0._wp
     real(wp) :: q_max = 0._wp
     real(wp) :: alpha = 0._wp
     real(wp) :: eps = 0._wp
     real(wp), allocatable :: q_in
     integer :: order
     integer :: log_order
     integer :: nlep
     logical :: running
   contains
     procedure :: init => qed_pdf_init
     procedure :: write => qed_pdf_write
     procedure :: set_order => qed_pdf_set_order
     procedure :: evolve_qed_pdf => qed_pdf_evolve_qed_pdf
     procedure :: allocate_aqed => qed_pdf_allocate_aqed
  end type qed_pdf_t


  interface
    module pure function coeffqed_b0(nf, nlep)
      integer, intent(in) :: nf, nlep
      real(wp) :: n_lep, coeffqed_b0
    end function coeffqed_b0
    module pure function coeffqed_b1(nf, nlep)
      integer, intent(in) :: nf, nlep
      real(wp) :: n_lep, coeffqed_b1
    end function coeffqed_b1
    module subroutine qed_pdf_init &
         (qed_pdf, mass, alpha, charge, q_max, order, log_order, nlep)
      class(qed_pdf_t), intent(out) :: qed_pdf
      real(default), intent(in) :: mass, alpha, q_max, charge
      integer, intent(in) :: order, log_order, nlep
    end subroutine qed_pdf_init
    module subroutine qed_pdf_write (qed_pdf, unit, with_qed)
      class(qed_pdf_t), intent(in) :: qed_pdf
      integer, intent(in), optional :: unit
      logical, intent(in), optional :: with_qed
    end subroutine qed_pdf_write
    module subroutine qed_pdf_set_order (qed_pdf, order)
      class(qed_pdf_t), intent(inout) :: qed_pdf
      integer, intent(in) :: order
    end subroutine qed_pdf_set_order
    module subroutine qed_pdf_evolve_qed_pdf (qed_pdf, x, xb, rb, ff)
      class(qed_pdf_t), intent(inout) :: qed_pdf
      real(default), intent(in) :: x, xb, rb
      real(default), intent(inout) :: ff
    end subroutine qed_pdf_evolve_qed_pdf
    module subroutine set_qed_pdf_parameters &
        (epdf, scale, alpha, running, nlep, nf, ln0, eta0, p, al0_2pi, al_2pi, n, run, order)
      type(qed_pdf_t), intent(in) :: epdf
      real(wp), intent(in) :: scale
      logical, intent(out) :: running
      real(wp), intent(out) :: ln0, eta0, p, alpha, al0_2pi, al_2pi, n, run
      integer, intent(out) :: nlep, nf
      logical, dimension(6), intent(out) :: order
    end subroutine set_qed_pdf_parameters
    module pure function elec_asym &
        (xb, nlep, nf, n, p, al0_2pi, al_2pi, order, running) &
        result (elec_as)
      real(wp), dimension(7,7,2) :: elec_as
      real(wp), intent(in) :: xb
      real(wp), intent(in) :: n, p, al0_2pi, al_2pi
      integer, intent(in) :: nf, nlep
      logical, intent(in) :: running
      logical, intent(in), dimension(6) :: order
    end function elec_asym
    module pure function phot_asym &
         (x, xb, log_xb, n, p, al0_2pi, order, running) result (phot_as)
      real(wp), dimension(7,7,2) :: phot_as
      real(wp), intent(in) :: x, xb, log_xb
      real(wp), intent(in) :: p, n, al0_2pi
      logical, intent(in) :: running
      logical, intent(in), dimension(6) :: order
    end function phot_asym
    module function recasym &
         (flv, x, xb, log_xb, nlep, nf, n, p, &
         al0_2pi, al_2pi, order, running) result (jll_nll)
      integer, intent(in) :: flv, nlep, nf
      real(wp), intent(in) :: x, xb, log_xb
      real(wp), intent(in) :: n, p, al0_2pi, al_2pi
      logical, intent(in) :: running
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2) :: jll_nll
    end function recasym
    module function bar_asym (flv, xb, n, run, order) result (jll_nll)
      integer, intent(in) :: flv
      real(wp), intent(in) :: xb
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2) :: jll_nll
    end function bar_asym
    module function recnum &
         (flv, x, xb, xp, pl, n) result (jll_nll)
      integer, intent(in) :: flv
      real(wp), dimension(7,7,2) :: jll_nll
      real(wp), intent(in) :: x, xb, xp, n
      real(wp), dimension(12), intent(in) :: pl
    end function recnum
    module function recbar &
         (flv, x, xb, n, run, order) result (jll_nll)
      integer, intent(in) :: flv
      real(wp), intent(in) :: x, xb
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2) :: jll_nll
    end function recbar
    module function rechat &
         (flv, x, xb, xp, pl, n, run, order) result (jll_nll)
      integer, intent(in) :: flv
      real(wp), intent(in) :: x, xb, xp
      real(wp), dimension(12), intent(in) :: pl
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2) :: jll_nll
    end function rechat
    module subroutine elecbar_asym_p (xb, jll_nll, n, run, order)
      real(wp), intent(in) :: xb
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine elecbar_asym_p
    module subroutine photbar_asym_p (jll_nll, n, run, order)
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine photbar_asym_p
    module subroutine recbar_singlet (x, xb, jll_nll, n, run, order)
      real(wp), intent(in) :: x, xb
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine recbar_singlet
    module subroutine recbar_nonsinglet (x, xb, jll_nll, n, run, order)
      real(wp), intent(in) :: x, xb
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine recbar_nonsinglet
    module subroutine recbar_photon (jll_nll, x, xb, n, run, order)
      real(wp), intent(in) :: x, xb
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine recbar_photon
    module subroutine recbar_ele (x, xb, jll_nll, n, run, order)
      real(wp), intent(in) :: x, xb
      real(wp), intent(in) ::  n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine recbar_ele
    module subroutine recbar_pos (jll_nll, n, run, order)
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine recbar_pos
    module subroutine rechat_singlet (x, xb, xp, pl, jll_nll, n, run, order)
      real(wp), intent(in) :: x, xb, xp
      real(wp), dimension(12), intent(in) :: pl
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine rechat_singlet
    module subroutine rechat_nonsinglet (x, xb, xp, pl, jll_nll, n, run, order)
      real(wp), intent(in) :: x, xb, xp
      real(wp), dimension(12), intent(in) :: pl
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine rechat_nonsinglet
    module subroutine rechat_photon (x, xb, xp, pl, jll_nll, n, run, order)
      real(wp), intent(in) :: x, xb, xp
      real(wp), dimension(12), intent(in) :: pl
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine rechat_photon
    module subroutine rechat_ele (x, xb, xp, pl, jll_nll, n, run, order)
      real(wp), intent(in) :: x, xb, xp
      real(wp), dimension(12), intent(in) :: pl
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine rechat_ele
    module subroutine rechat_pos (x, xb, xp, pl, jll_nll, n, run, order)
      real(wp), intent(in) :: x, xb, xp
      real(wp), dimension(12), intent(in) :: pl
      real(wp), intent(in) :: n, run
      logical, dimension(6), intent(in) :: order
      real(wp), dimension(7,7,2), intent(out) :: jll_nll
    end subroutine rechat_pos
    module function t_alpha (epdf, scale) result (t)
      real(wp) :: t
      type(qed_pdf_t), intent(in) :: epdf
      real(wp), intent(in) :: scale
    end function t_alpha
   module pure function full_series ( &
       log_x, log_xb, ln0, p, al_2pi, a01, a02, a03, a04, a05) result (res)
     real(wp) :: res
     real(wp), intent(in) :: log_x, log_xb, p, al_2pi, ln0
     real(wp), intent(in), dimension(7,7,2) :: a01
     real(wp), intent(in), dimension(7,7,2), optional :: a02, a03, a04, a05
   end function full_series
   module pure function rec_log_series (log_x, log_xb, ln0, expansion) result (jll_nll)
     real(wp), intent(in) :: log_x, log_xb, ln0
     real(wp), dimension(7,7,2), intent(in) :: expansion
     real(wp), dimension(6) :: jll_nll
   end function rec_log_series
   module pure function add_logvec (expansion1, expansion2) result (jll_nll)
     real(wp), dimension(7,7,2), intent(in) :: expansion1, expansion2
     real(wp), dimension(7,7,2) :: jll_nll
   end function add_logvec
   module pure function sub_logvec (expansion1, expansion2) result (jll_nll)
     real(wp), dimension(7,7,2), intent(in) :: expansion1, expansion2
     real(wp), dimension(7,7,2) :: jll_nll
   end function sub_logvec
    module pure function endpoint_func_S (x, xb, xp, pl, n) result (ints)
      real(wp), intent(in) :: x, xb, xp, n
      real(wp), dimension(12), intent(in) :: pl
      real(wp), dimension(7) :: ints
    end function endpoint_func_S
    module pure function endpoint_func_NS (x, xb, xp, pl, n) result (ints)
      real(wp), intent(in) :: x, xb, xp, n
      real(wp), dimension(12), intent(in) :: pl
      real(wp), dimension(7) :: ints
    end function endpoint_func_NS
    module pure function endpoint_func_GAM (x, xb, xp, pl) result (ints)
      real(wp), intent(in) :: x, xb, xp
      real(wp), dimension(12), intent(in) :: pl
      real(wp), dimension(7) :: ints
    end function endpoint_func_GAM
    module pure function endpoint_func_POS (x, xb, xp, pl, n) result (ints)
      real(wp), intent(in) :: x, xb, xp, n
      real(wp), dimension(12), intent(in) :: pl
      real(wp), dimension(7) :: ints
    end function endpoint_func_POS
    module pure function endpoint_func_ELE (x, xb, xp, pl, n) result (ints)
      real(wp), intent(in) :: x, xb, xp, n
      real(wp), dimension(12), intent(in) :: pl
      real(wp), dimension(7) :: ints
    end function endpoint_func_ELE
    module function elec_pdf (epdf, flv, x, xb, scale) result (res)
      type(qed_pdf_t), intent(in) :: epdf
      real(wp) :: res
      integer, intent(in) :: flv
      real(wp), intent(in) :: x, xb
      real(wp), intent(in) :: scale
    end function elec_pdf
  end interface

contains

  subroutine qed_pdf_allocate_aqed (qed, order, n_f, nlep, running)
    class(qed_pdf_t), intent(inout) :: qed
    integer, intent(in) :: order, n_f, nlep
    logical, intent(in) :: running
    qed%running = running
    if (running) then
       allocate (alpha_qed_from_scale_t :: qed%aqed)
       select type (aqed => qed%aqed)
       type is (alpha_qed_from_scale_t)
          aqed%order = order
          aqed%nf = n_f
          aqed%nlep = nlep
       end select
    else
       allocate (alpha_qed_fixed_t :: qed%aqed)
    end if
  end subroutine qed_pdf_allocate_aqed


end module electron_pdfs
